#include "GrTown_PCH.H"
#include "Objects.H"
#include "DrawUtils.H"
#include "Texture.H"
#include "MMatrix.H"
#include "Houses.H"
#include "Roads.H"

// some very basic classes of objects...
// mainly for debugging
Cube::Cube(float x, float y, float z, float s, 
		   float r, float g, float b)
  : color(r,g,b), size(s)
{
  transMatrix(transform,x,y,z);
}
void Cube::draw(DrawingState*)
{
  glColor4fv(&color.r);
  cube(0,0,0,size);
}

/////////////////////////////////////////
// stuff for all houses
static int houseColors[][3] = {
	{240,240,240}, {180,175,100}, 
	{200,100,100}, {147,144,244},
	{250,249,157}, {199,144,186}
};
int nHouseColors = 6;


/***********************************************************************/
TestCubes::TestCubes()
{
}

void TestCubes::draw(DrawingState*)
{
  glColor3f(1,0,0);
  cube( 20,0,0, 3);
  cube(-20,0,0, 3);
  cube(0, 20,0, 3);
  cube(0,-20,0, 3);
  cube(0,0, 20, 3);
  cube(0,0,-20, 3);

  glColor3f(1,1,0);
  cube(20,10,20,3);
  glColor3f(0,1,1);
  cube(-20,10,20,3);
  glColor3f(1,0,1);
  cube(-20,10,-20,3);
  glColor3f(1,1,1);
  cube(20,10,-20,3);
}

/***********************************************************************/
// a Really simple tree - just a cone with a stem...
// use normals to make it look rounder...
SimpleTree1::SimpleTree1(float h, float r)
  : height(h), radius(r)
{
}
void SimpleTree1::draw(DrawingState*)
{
  // cone for the body
  glFrontFace(GL_CW);
  glColor3f(0.f,.6f,.3f);
  glBegin(GL_TRIANGLE_FAN);
  glNormal3f(0,1,0);
  glVertex3f(0,height,0);
  glNormal3f(1,0,0);
  glVertex3f(radius,height/4,0);
  glNormal3f(0,0,1);
  glVertex3f(0,height/4,radius);
  glNormal3f(-1,0,0);
  glVertex3f(-radius,height/4,0);
  glNormal3f(0,0,-1);
  glVertex3f(0,height/4,-radius);
  glNormal3f(1,0,0);
  glVertex3f(radius,height/4,0);
  glEnd();
  glFrontFace(GL_CCW);
  // stem
  glColor3f(.6f,.5f,.3f);
  glBegin(GL_QUAD_STRIP);
  glNormal3f(1,0,0);
  glVertex3f(radius/5,0,0);
  glVertex3f(radius/5,height/4,0);
  glNormal3f(0,0,1);
  glVertex3f(0,0,radius/5);
  glVertex3f(0,height/4,radius/5);
  glNormal3f(-1,0,0);
  glVertex3f(-radius/5,0,0);
  glVertex3f(-radius/5,height/4,0);
  glNormal3f(0,0,-1);
  glVertex3f(0,0,-radius/5);
  glVertex3f(0,height/4,-radius/5);
  glNormal3f(1,0,0);
  glVertex3f(radius/5,0,0);
  glVertex3f(radius/5,height/4,0);
  glEnd();
}
/**********************************************************************/
Sign::Sign(float sxi, float syi, float hi, char* tex, 
		   SignShape ss)
  : sx(sxi), sy(syi), h(hi), texName(tex), shape(ss)
{
}
void Sign::draw(DrawingState*)
{
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE, GL_MODULATE);
  fetchTexture(texName);
  glBegin(GL_POLYGON);
  glNormal3f(0,0,1);
  glColor3f(1,1,1);
  drawShape(0);
  glEnd();
  // back of the sign
  glBindTexture(GL_TEXTURE_2D,0);
  glNormal3f(0,0,-1);
  drawShape(-.1f);
  // post
  glBegin(GL_POLYGON);
    glNormal3f(0,0,1);
	glVertex3f(-.1f,h,0);
	glVertex3f( .1f,h,0);
	glVertex3f( .1f,0,0);
	glVertex3f(-.1f,0,0);
  glEnd();
  glFrontFace(GL_CW);
  glBegin(GL_POLYGON);
    glNormal3f(0,0,1);
	glVertex3f(-.1f,h,-.05f);
	glVertex3f( .1f,h,-.05f);
	glVertex3f( .1f,0,-.05f);
	glVertex3f(-.1f,0,-.05f);
  glEnd();
  glFrontFace(GL_CCW);	
}
void Sign::drawShape(float z)
{
  glFrontFace(GL_CW);
  switch(shape) {
  case rectangle:
	glBegin(GL_POLYGON);
	glTexCoord2f(0.,0.); glVertex3f(-sx/2,h,z);
	glTexCoord2f(1.,0.); glVertex3f( sx/2,h,z);
	glTexCoord2f(1.,1.); glVertex3f( sx/2,h+sy,z);
	glTexCoord2f(0.,1.); glVertex3f(-sx/2,h+sy,z);
	glEnd();
	break;
  case octagon:
	glBegin(GL_POLYGON);
	glTexCoord2f(0.25f,0.f); glVertex3f(-sx/4,h,z);
	glTexCoord2f(0.75f,0.f); glVertex3f( sx/4,h,z);
	glTexCoord2f(1.f,.25f); glVertex3f( sx/2.f,h+    sy/4.f,z);
	glTexCoord2f(1.f,.75f); glVertex3f( sx/2.f,h+3.f*sy/4.f,z);
	glTexCoord2f(.75f,1); glVertex3f( sx/4,h+sy,z);
	glTexCoord2f(.25f,1); glVertex3f(-sx/4,h+sy,z);
	glTexCoord2f(0,.75f); glVertex3f(-sx/2.f,h+3.f*sy/4.f,z);
	glTexCoord2f(0,.25f); glVertex3f(-sx/2.f,h+sy/4.f,z);
	glEnd();
	break;
  case diamond:
	glBegin(GL_POLYGON);
	glTexCoord2f(0.45f, 0.f); glVertex3f(-sx/10,h,z);
	glTexCoord2f(0.55f, 0.f); glVertex3f( sx/10,h,z);
	glTexCoord2f(   1 ,.45f); glVertex3f( sx/2,h+9*sy/20.f,z);
	glTexCoord2f(   1 ,.55f); glVertex3f( sx/2,h+11*sy/20.f,z);
	glTexCoord2f(.55f, 1   ); glVertex3f( sx/10,h+sy,z);
	glTexCoord2f(.45f, 1   ); glVertex3f(-sx/10,h+sy,z);
	glTexCoord2f(   0 ,.55f); glVertex3f(-sx/2,h+11*sy/20.f,z);
	glTexCoord2f(   0 ,.45f); glVertex3f(-sx/2,h+ 9*sy/20.f,z);
	glEnd();
	break;
  };
  glFrontFace(GL_CCW);
}

/***********************************************************************/
// draw a patch of lawn - use a repeating texture
Lawn::Lawn(float xi1, float zi1, float xi2, float zi2)
  : x1(xi1), z1(zi1), x2(xi2), z2(zi2)
{
}

void Lawn::draw(DrawingState* d)
{
  // the catch here is that we need to use a polygon offset to draw
  // the lawn just above the ground...
  if (d->drGrTex)
	fetchTexture("grass",true,true);
  else
    glBindTexture(GL_TEXTURE_2D,0);
  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(-2.,-2.);
  glNormal3f(0,1,0);
  glColor3f(0,1,1);
  glBegin(GL_POLYGON);
  glTexCoord2f(0,(z2-z1)/4.f);            glVertex3f(x1,0,z2);
  glTexCoord2f((x2-x1)/4.f, (z2-z1)/4.f); glVertex3f(x2,0,z2);
  glTexCoord2f((x2-x1)/4.f,0);            glVertex3f(x2,0,z1);
  glTexCoord2f(0,0);                      glVertex3f(x1,0,z1);
  glEnd();
  glDisable(GL_POLYGON_OFFSET_FILL);
  glBindTexture(GL_TEXTURE_2D,0);
}

// draw a little suburban dream...
// a 100x200 lot...
SimpleLot::SimpleLot(int ht, int hc)  
{
  // make the things we want
  add(new Lawn(0,0,100,200),0,0,0,0);
  switch(ht) {
  case 0: {
	  SimpleHouse1* h1 = new SimpleHouse1();
	  h1->color(houseColors[hc]);
	  add(h1,50.,0,60.,0);
	  break; }
  case 1: {
	  SimpleHouse2* h2 = new SimpleHouse2();
	  h2->color(houseColors[hc]);
	  add(h2,50.,0,60.,0);
	  break; }
  case 2: {
	  SimpleHouse3* h3 = new SimpleHouse3();
	  h3->color(houseColors[hc]);
	  add(h3,50.,0,60.,0);
	  break; }
  } 
  add(new SimpleTree1(15,5), 10,0, 10,0);
  add(new SimpleTree1(15,5), 10,0,190,0);
  add(new SimpleTree1(15,5), 30,0,190,0);
  add(new SimpleTree1(15,5), 50,0,190,0);
  add(new SimpleTree1(15,5), 70,0,190,0);
}

/***********************************************************************/
Road::Road()
{
	// install onto linked list
//	nextRoad = theRoads;
//	theRoads = this;
}
void Road::draw(DrawingState*)
{
  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(-1.,-2.);
  glColor3f(.6f,.6f,.6f);
  glNormal3f(0,1,0);
  fetchTexture("asphalt",true,true);
  roadDraw();

  glDisable(GL_POLYGON_OFFSET_FILL);

}
Road* theRoads = 0;

StraightRoad::StraightRoad(int d, int l, int nl, int s)
  : length(l), nLanes(nl), shoulders(s)
{
	rotMatrix(transform, 'Y', (float) (((float) (d*90)) * 3.141592f / 180.f));
}

void StraightRoad::roadDraw()
{
  glBegin(GL_POLYGON);
  float rw = (float) (15*nLanes + shoulders*10) / 2.f;
  glTexCoord2f(0,(float)length/15.f);		   glVertex3f(-rw,0,(float)length);
  glTexCoord2f((float) nLanes,(float) length/15.f);    glVertex3f(rw,0,(float)length);
  glTexCoord2f((float) nLanes,0);				   glVertex3f(rw,0,0);
  glTexCoord2f(0,0);                       glVertex3f(-rw,0,0);
  glEnd();
}


/***********************************************************************/
SimpleSubdivision::SimpleSubdivision(int nh) 
{
  add(new StreetLight(7), 50,0, -5,0);
  add(new StreetLight(-7),150,0,-25,0);
  add(new StreetLight(7),250,0, -5,0);
  add(new StreetLight(-7),350,0,-25,0);
  add(new StreetLight(7),450,0, -5,0);
  add(new StreetLight(-7),550,0,-25,0);
  for (int hc=0; hc<nh; hc++) {
	add(new SimpleLot(rand()%3,rand() % nHouseColors), (float) (hc*100), 0, 0, 0);
	GrObject* g2 = new SimpleLot((hc+1)%2+1,(20-hc)%3);
	rotMatrix(g2->transform,'Y',3.14159f);
	add(g2,(float) (hc*100+100),0,-30,0);
  }
  add(new StraightRoad(1,100*nh,2),0,0,-15,0);
  add(new Sign(4,4,4,"stop",octagon),(float)(100*nh-2),0,2,-(3.141592f/2.f));
  add(new Sign(4,4,4,"stop",octagon),  2,0,-32,(3.141592f/2.f));
  add(new Sign(4,4,4,"stop",octagon),(float)(100*nh-2),0,-32,(3.141592f));
  add(new Sign(4,4,4,"stop",octagon),  2,0,2,0);

}

SubdivisionBlocks::SubdivisionBlocks(int nr, int nc, int nh) 
  : nRows(nr), nCols(nc)
{
  int r,c,x0,z0;
  int blockSizeX = nh*100 + 30;
  int x1 = 15;
  int z1 = 15;
  int origin = -1;
  for(z0=z1,r = 0; r<nr; r++, z0 += 430) {
	for(c = 0,x0=x1; c<nc; c++, x0 += blockSizeX) {
	  Intersection* i = new Intersection(2);
	  if (origin<0) origin = i->id-1;
	  add(new SimpleSubdivision(nh),(float) x0,0,(float) z0,0);
	  add(new StraightRoad(2,200),(float)(x0-15),0,(float)z0-30,0);
	  add(new StraightRoad(2,200),(float)(x0-15),0,(float)z0+200,0);
	  add(i,(float)(x0-15),0,(float)(z0-15),0);
	}
  }
  printf("There are %d Intersections!\n",theIntersectionsV.size());
  // wire together the intersections
  // north and sound bound

  for(r=0; r<nr-1; r++)
	for(c=0; c<nc; c++) {
	  theIntersectionsV[origin+r*nc+c]->neighbor[1] = origin+(r+1)*nc+c+1;
	  theIntersectionsV[origin+(r+1)*nc+c]->neighbor[3] = origin+(r)*nc+c+1;
	}
  // the westbound and eastbound roads
  for(r=0; r<nr; r++)
	for(c=0; c<nc-1; c++) {
	  theIntersectionsV[origin+r*nc+c]->neighbor[0] = origin+r*nc+c+1+1;
	  theIntersectionsV[origin+r*nc+c+1]->neighbor[2] = origin+(r)*nc+c+1;
	}
}

/*&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&*/
StreetLight::StreetLight(double p) : poleZ(p)
{
	radius = 10;
}
void StreetLight::draw(DrawingState*)
{
  glColor3f(.6f,.5f,.3f);
  glBegin(GL_QUAD_STRIP);
  glNormal3d( 1, 0,0);
  glVertex3d( 1, 0,poleZ);
  glVertex3d( 1,20,poleZ);
  glNormal3d( 0, 0,1);
  glVertex3d( 0, 0,poleZ+1);
  glVertex3d( 0,20,poleZ+1);
  glNormal3d(-1, 0,0);
  glVertex3d(-1, 0,poleZ);
  glVertex3d(-1,20,poleZ);
  glNormal3d( 0, 0,-1);
  glVertex3d( 0, 0,poleZ-1);
  glVertex3d( 0,20,poleZ-1);
  glNormal3d( 1, 0,0);
  glVertex3d( 1, 0,poleZ);
  glVertex3d( 1,20,poleZ);
  glEnd();
  glBegin(GL_QUAD_STRIP);
  glNormal3d(0,1,0);
  glVertex3d( 0,21, 0);
  glVertex3d( 0,21, poleZ);
  glNormal3d(-.707,-.707, 0);
  glVertex3d(-1,20, 0);
  glVertex3d(-1,20, poleZ);
  glNormal3d(.707,-.707, 0);
  glVertex3d( 1,20, 0);
  glVertex3d( 1,20, poleZ);
  glNormal3d(0,1,0);
  glVertex3d( 0,21, 0);
  glVertex3d( 0,21, poleZ);
	glEnd();
}

void StreetLight::drawAfter(DrawingState* d)
{
  int tod = d->timeOfDay;
  bool daytime;
  if (tod >=6 && tod <= 19)
	daytime = true;
  else 
	daytime = false;

  if (!daytime) {

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_DST_ALPHA);
	double cone_length = 20;
	double radius = 10;
	double theta,dtheta = PI/16;
	int c;

	double norm[3];
	double sum;
	// Beam One
	glPushMatrix();
			
	glFrontFace(GL_CW);
			
	glBegin(GL_TRIANGLE_FAN);
	glColor4d(1,1,0,.8);
	glVertex3d(0,cone_length,0);
	for ( c=0,theta = 0.0 ; c <= 32  ; theta += dtheta, c++ ) {
	  double z = sin(theta) * radius;
	  double x = cos(theta) * radius;
	  glColor4d(1,1,0,.2);
	
	  norm[0] = 0; norm[1] = x; norm[2] = z; sum = norm[0] + norm[1] + norm[2];
	  norm[0] /= sum; norm[1] /= sum; norm[2] /= sum;				
	  glNormal3d( norm[0],norm[1],norm[2]);
	
	  glVertex3d( x, cone_length/6, z );
	}  
			
	glEnd();

			
	glPopMatrix();

	glDisable(GL_BLEND);
	glClearColor(1,1,1,1);

  }
}

/***********************************************************************/
Intersection::Intersection(int nl) : nLanes(nl)
{
  theIntersectionsV.push_back(this);
  id = theIntersectionsV.size();
  cachedPosition = false;
  neighbor[0] = 0;
  neighbor[1] = 0;
  neighbor[2] = 0;
  neighbor[3] = 0;
}
void Intersection::roadDraw()
{
  float l = nLanes * 2.f;
  glBegin(GL_POLYGON);
  glTexCoord2f(0,l);     glVertex3i(-(nLanes/2)*15,0, (nLanes/2)*15);
  glTexCoord2f(l,l);     glVertex3i( (nLanes/2)*15,0, (nLanes/2)*15);
  glTexCoord2f(l,0);     glVertex3i( (nLanes/2)*15,0,-(nLanes/2)*15);
  glTexCoord2f(0,0);     glVertex3i(-(nLanes/2)*15,0,-(nLanes/2)*15);
  glEnd();
}
void Intersection::getPosition(float& x, float& z)
{
  if (!cachedPosition) {
	Matrix carry;
	Matrix tmp;

	copyMatrix(transform,carry);
	GrObject* p = parent;

	while(p) {
	  multMatrix(p->transform,carry,tmp);
	  copyMatrix(tmp,carry);
	  p = p->parent;
	}

	pcacheX = carry[3][0];
	pcacheZ = carry[3][2];
	cachedPosition = true;
  }
  x = pcacheX;
  z = pcacheZ;
}
IntersectionVector theIntersectionsV;

// assume its unit vector in compass direction
int direction(float x, float z )
{
  if (x>.5) return 0;
  else if (z>.5) return 1;
  else if (x<-.5) return 2;
  else if (z<-.5) return 3;
  else {
	printf("Bad Direction Vector! %g %g\n",x,z);
	return 0;
  }
}
void direction(int d, float& x, float& z)
{
	if (d<0) d+=4;
	d = d%4;
	switch(d) {
	case 0:  x =  1;	z =  0; break;
	case 1:  x =  0;	z =  1; break;
	case 2:  x = -1;	z =  0; break;
	case 3:  x =  0;	z = -1; break;
	};
}
